@@ -10,7 +10,8 @@ import warnings
 from collections.abc import Container, Iterable, Sized
 from pathlib import Path
 from types import MappingProxyType
-from urllib.parse import unquote, urlencode
+
+from yarl import URL, quote, unquote
 
 from . import hdrs
 from .abc import AbstractMatchInfo, AbstractRouter, AbstractView
@@ -22,9 +23,8 @@ from .web_reqrep import Response, StreamResponse
 
 __all__ = ('UrlDispatcher', 'UrlMappingMatchInfo',
            'AbstractResource', 'Resource', 'PlainResource', 'DynamicResource',
-           'ResourceAdapter',
            'AbstractRoute', 'ResourceRoute',
-           'Route', 'PlainRoute', 'DynamicRoute', 'StaticRoute', 'View')
+           'StaticResource', 'View')
 
 
 PY_35 = sys.version_info >= (3, 5)
@@ -44,26 +44,38 @@ class AbstractResource(Sized, Iterable):
 
     @abc.abstractmethod  # pragma: no branch
     def url(self, **kwargs):
+        """Construct url for resource with additional params.
+
+        Deprecated, use url_for() instead.
+
+        """
+        warnings.warn(".url(...) is deprecated, use .url_for instead",
+                      DeprecationWarning,
+                      stacklevel=3)
+
+    @abc.abstractmethod  # pragma: no branch
+    def url_for(self, **kwargs):
         """Construct url for resource with additional params."""
 
     @asyncio.coroutine
     @abc.abstractmethod  # pragma: no branch
-    def resolve(self, method, path):
+    def resolve(self, request):
         """Resolve resource
 
         Return (UrlMappingMatchInfo, allowed_methods) pair."""
 
+    @abc.abstractmethod
+    def add_prefix(self, prefix):
+        """Add a prefix to processed URLs.
+
+        Required for subapplications support.
+
+        """
+
     @abc.abstractmethod
     def get_info(self):
         """Return a dict with additional info useful for introspection"""
 
-    @staticmethod
-    def _append_query(url, query):
-        if query:
-            return url + "?" + urlencode(query)
-        else:
-            return url
-
 
 class AbstractRoute(abc.ABC):
 
@@ -127,9 +139,20 @@ class AbstractRoute(abc.ABC):
         """Return a dict with additional info useful for introspection"""
 
     @abc.abstractmethod  # pragma: no branch
-    def url(self, **kwargs):
+    def url_for(self, *args, **kwargs):
         """Construct url for route with additional params."""
 
+    @abc.abstractmethod  # pragma: no branch
+    def url(self, **kwargs):
+        """Construct url for resource with additional params.
+
+        Deprecated, use url_for() instead.
+
+        """
+        warnings.warn(".url(...) is deprecated, use .url_for instead",
+                      DeprecationWarning,
+                      stacklevel=3)
+
     @asyncio.coroutine
     def handle_expect_header(self, request):
         return (yield from self._expect_handler(request))
@@ -140,6 +163,8 @@ class UrlMappingMatchInfo(dict, AbstractMatchInfo):
     def __init__(self, match_dict, route):
         super().__init__(match_dict)
         self._route = route
+        self._apps = []
+        self._frozen = False
 
     @property
     def handler(self):
@@ -160,6 +185,18 @@ class UrlMappingMatchInfo(dict, AbstractMatchInfo):
     def get_info(self):
         return self._route.get_info()
 
+    @property
+    def apps(self):
+        return tuple(self._apps)
+
+    def add_app(self, app):
+        if self._frozen:
+            raise RuntimeError("Cannot change apps stack after .freeze() call")
+        self._apps.insert(0, app)
+
+    def freeze(self):
+        self._frozen = True
+
     def __repr__(self):
         return "<MatchInfo {}: {}>".format(super().__repr__(), self._route)
 
@@ -194,40 +231,6 @@ def _defaultExpectHandler(request):
             raise HTTPExpectationFailed(text="Unknown Expect: %s" % expect)
 
 
-class ResourceAdapter(AbstractResource):
-
-    def __init__(self, route):
-        assert isinstance(route, Route), \
-            'Instance of Route class is required, got {!r}'.format(route)
-        super().__init__(name=route.name)
-        self._route = route
-        route._resource = self
-
-    def url(self, **kwargs):
-        return self._route.url(**kwargs)
-
-    @asyncio.coroutine
-    def resolve(self, method, path):
-        route_method = self._route.method
-        allowed_methods = set()
-        match_dict = self._route.match(path)
-        if match_dict is not None:
-            allowed_methods.add(route_method)
-            if route_method == hdrs.METH_ANY or route_method == method:
-                return (UrlMappingMatchInfo(match_dict, self._route),
-                        allowed_methods)
-        return None, allowed_methods
-
-    def get_info(self):
-        return self._route.get_info()
-
-    def __len__(self):
-        return 1
-
-    def __iter__(self):
-        yield self._route
-
-
 class Resource(AbstractResource):
 
     def __init__(self, *, name=None):
@@ -254,10 +257,10 @@ class Resource(AbstractResource):
         self._routes.append(route)
 
     @asyncio.coroutine
-    def resolve(self, method, path):
+    def resolve(self, request):
         allowed_methods = set()
 
-        match_dict = self._match(path)
+        match_dict = self._match(request.rel_url.raw_path)
         if match_dict is None:
             return None, allowed_methods
 
@@ -265,7 +268,7 @@ class Resource(AbstractResource):
             route_method = route.method
             allowed_methods.add(route_method)
 
-            if route_method == method or route_method == hdrs.METH_ANY:
+            if route_method == request.method or route_method == hdrs.METH_ANY:
                 return UrlMappingMatchInfo(match_dict, route), allowed_methods
         else:
             return None, allowed_methods
@@ -281,8 +284,15 @@ class PlainResource(Resource):
 
     def __init__(self, path, *, name=None):
         super().__init__(name=name)
+        assert path.startswith('/')
         self._path = path
 
+    def add_prefix(self, prefix):
+        assert prefix.startswith('/')
+        assert prefix.endswith('/')
+        assert len(prefix) > 1
+        self._path = prefix + self._path[1:]
+
     def _match(self, path):
         # string comparison is about 10 times faster than regexp matching
         if self._path == path:
@@ -294,7 +304,11 @@ class PlainResource(Resource):
         return {'path': self._path}
 
     def url(self, *, query=None):
-        return self._append_query(self._path, query)
+        super().url()
+        return str(self.url_for().with_query(query))
+
+    def url_for(self):
+        return URL(self._path)
 
     def __repr__(self):
         name = "'" + self.name + "' " if self.name is not None else ""
@@ -306,11 +320,20 @@ class DynamicResource(Resource):
 
     def __init__(self, pattern, formatter, *, name=None):
         super().__init__(name=name)
+        assert pattern.pattern.startswith('\\/')
+        assert formatter.startswith('/')
         self._pattern = pattern
         self._formatter = formatter
 
+    def add_prefix(self, prefix):
+        assert prefix.startswith('/')
+        assert prefix.endswith('/')
+        assert len(prefix) > 1
+        self._pattern = re.compile(re.escape(prefix)+self._pattern.pattern[2:])
+        self._formatter = prefix + self._formatter[1:]
+
     def _match(self, path):
-        match = self._pattern.match(path)
+        match = self._pattern.fullmatch(path)
         if match is None:
             return None
         else:
@@ -321,9 +344,13 @@ class DynamicResource(Resource):
         return {'formatter': self._formatter,
                 'pattern': self._pattern}
 
-    def url(self, *, parts, query=None):
+    def url_for(self, **parts):
         url = self._formatter.format_map(parts)
-        return self._append_query(url, query)
+        return URL(url)
+
+    def url(self, *, parts, query=None):
+        super().url(**parts)
+        return str(self.url_for(**parts).with_query(query))
 
     def __repr__(self):
         name = "'" + self.name + "' " if self.name is not None else ""
@@ -331,120 +358,30 @@ class DynamicResource(Resource):
                 .format(name=name, formatter=self._formatter))
 
 
-class ResourceRoute(AbstractRoute):
-    """A route with resource"""
+class PrefixResource(AbstractResource):
 
-    def __init__(self, method, handler, resource, *,
-                 expect_handler=None):
-        super().__init__(method, handler, expect_handler=expect_handler,
-                         resource=resource)
-
-    def __repr__(self):
-        return "<ResourceRoute [{method}] {resource} -> {handler!r}".format(
-            method=self.method, resource=self._resource,
-            handler=self.handler)
-
-    @property
-    def name(self):
-        return self._resource.name
-
-    def url(self, **kwargs):
-        """Construct url for route with additional params."""
-        return self._resource.url(**kwargs)
-
-    def get_info(self):
-        return self._resource.get_info()
-
-    _append_query = staticmethod(Resource._append_query)
-
-
-class Route(AbstractRoute):
-    """Old fashion route"""
-
-    def __init__(self, method, handler, name, *, expect_handler=None):
-        super().__init__(method, handler, expect_handler=expect_handler)
-        self._name = name
-
-    @property
-    def name(self):
-        return self._name
-
-    @abc.abstractmethod
-    def match(self, path):
-        """Return dict with info for given path or
-        None if route cannot process path."""
-
-    _append_query = staticmethod(Resource._append_query)
-
-
-class PlainRoute(Route):
-
-    def __init__(self, method, handler, name, path, *, expect_handler=None):
-        super().__init__(method, handler, name, expect_handler=expect_handler)
-        self._path = path
-
-    def match(self, path):
-        # string comparison is about 10 times faster than regexp matching
-        if self._path == path:
-            return {}
-        else:
-            return None
-
-    def url(self, *, query=None):
-        return self._append_query(self._path, query)
-
-    def get_info(self):
-        return {'path': self._path}
-
-    def __repr__(self):
-        name = "'" + self.name + "' " if self.name is not None else ""
-        return "<PlainRoute {name}[{method}] {path} -> {handler!r}".format(
-            name=name, method=self.method, path=self._path,
-            handler=self.handler)
-
-
-class DynamicRoute(Route):
-
-    def __init__(self, method, handler, name, pattern, formatter, *,
-                 expect_handler=None):
-        super().__init__(method, handler, name, expect_handler=expect_handler)
-        self._pattern = pattern
-        self._formatter = formatter
-
-    def match(self, path):
-        match = self._pattern.match(path)
-        if match is None:
-            return None
-        else:
-            return match.groupdict()
-
-    def url(self, *, parts, query=None):
-        url = self._formatter.format_map(parts)
-        return self._append_query(url, query)
-
-    def get_info(self):
-        return {'formatter': self._formatter,
-                'pattern': self._pattern}
-
-    def __repr__(self):
-        name = "'" + self.name + "' " if self.name is not None else ""
-        return ("<DynamicRoute {name}[{method}] {formatter} -> {handler!r}"
-                .format(name=name, method=self.method,
-                        formatter=self._formatter, handler=self.handler))
-
-
-class StaticRoute(Route):
-
-    def __init__(self, name, prefix, directory, *,
-                 expect_handler=None, chunk_size=256*1024,
-                 response_factory=StreamResponse,
-                 show_index=False):
+    def __init__(self, prefix, *, name=None):
         assert prefix.startswith('/'), prefix
         assert prefix.endswith('/'), prefix
-        super().__init__(
-            'GET', self.handle, name, expect_handler=expect_handler)
-        self._prefix = prefix
+        super().__init__(name=name)
+        self._prefix = quote(prefix, safe='/')
         self._prefix_len = len(self._prefix)
+
+    def add_prefix(self, prefix):
+        assert prefix.startswith('/')
+        assert prefix.endswith('/')
+        assert len(prefix) > 1
+        self._prefix = prefix + self._prefix[1:]
+        self._prefix_len = len(self._prefix)
+
+
+class StaticResource(PrefixResource):
+
+    def __init__(self, prefix, directory, *, name=None,
+                 expect_handler=None, chunk_size=256*1024,
+                 response_factory=StreamResponse,
+                 show_index=False, follow_symlinks=False):
+        super().__init__(prefix, name=name)
         try:
             directory = Path(directory)
             if str(directory).startswith('~'):
@@ -459,30 +396,65 @@ class StaticRoute(Route):
         self._file_sender = FileSender(resp_factory=response_factory,
                                        chunk_size=chunk_size)
         self._show_index = show_index
+        self._follow_symlinks = follow_symlinks
+        self._expect_handler = expect_handler
 
-    def match(self, path):
-        if not path.startswith(self._prefix):
-            return None
-        return {'filename': path[self._prefix_len:]}
+        self._routes = {'GET': ResourceRoute('GET', self._handle, self,
+                                             expect_handler=expect_handler),
+
+                        'HEAD': ResourceRoute('HEAD', self._handle, self,
+                                              expect_handler=expect_handler)}
 
     def url(self, *, filename, query=None):
+        return str(self.url_for(filename=filename).with_query(query))
+
+    def url_for(self, *, filename):
         if isinstance(filename, Path):
             filename = str(filename)
         while filename.startswith('/'):
             filename = filename[1:]
-        url = self._prefix + filename
-        return self._append_query(url, query)
+        url = self._prefix + quote(filename, safe='/')
+        return URL(url)
 
     def get_info(self):
         return {'directory': self._directory,
                 'prefix': self._prefix}
 
+    def set_options_route(self, handler):
+        if 'OPTIONS' in self._routes:
+            raise RuntimeError('OPTIONS route was set already')
+        self._routes['OPTIONS'] = ResourceRoute(
+            'OPTIONS', handler, self,
+            expect_handler=self._expect_handler)
+
     @asyncio.coroutine
-    def handle(self, request):
+    def resolve(self, request):
+        path = request.rel_url.raw_path
+        method = request.method
+        allowed_methods = set(self._routes)
+        if not path.startswith(self._prefix):
+            return None, set()
+
+        if method not in allowed_methods:
+            return None, allowed_methods
+
+        match_dict = {'filename': unquote(path[self._prefix_len:])}
+        return (UrlMappingMatchInfo(match_dict, self._routes[method]),
+                allowed_methods)
+
+    def __len__(self):
+        return len(self._routes)
+
+    def __iter__(self):
+        return iter(self._routes.values())
+
+    @asyncio.coroutine
+    def _handle(self, request):
         filename = unquote(request.match_info['filename'])
         try:
             filepath = self._directory.joinpath(filename).resolve()
-            filepath.relative_to(self._directory)
+            if not self._follow_symlinks:
+                filepath.relative_to(self._directory)
         except (ValueError, FileNotFoundError) as error:
             # relatively safe
             raise HTTPNotFound() from error
@@ -495,7 +467,8 @@ class StaticRoute(Route):
         if filepath.is_dir():
             if self._show_index:
                 try:
-                    ret = Response(text=self._directory_as_html(filepath))
+                    ret = Response(text=self._directory_as_html(filepath),
+                                   content_type="text/html")
                 except PermissionError:
                     raise HTTPForbidden()
             else:
@@ -545,22 +518,104 @@ class StaticRoute(Route):
         return html
 
     def __repr__(self):
-        name = "'" + self.name + "' " if self.name is not None else ""
-        return "<StaticRoute {name}[{method}] {path} -> {directory!r}".format(
-            name=name, method=self.method, path=self._prefix,
-            directory=self._directory)
+        name = "'" + self.name + "'" if self.name is not None else ""
+        return "<StaticResource {name} {path} -> {directory!r}".format(
+            name=name, path=self._prefix, directory=self._directory)
 
 
-class SystemRoute(Route):
+class PrefixedSubAppResource(PrefixResource):
 
-    def __init__(self, http_exception):
-        super().__init__(hdrs.METH_ANY, self._handler, None)
-        self._http_exception = http_exception
+    def __init__(self, prefix, app):
+        super().__init__(prefix)
+        self._app = app
+        for resource in app.router.resources():
+            resource.add_prefix(prefix)
+
+    def add_prefix(self, prefix):
+        super().add_prefix(prefix)
+        for resource in self._app.router.resources():
+            resource.add_prefix(prefix)
+
+    def url_for(self, *args, **kwargs):
+        raise RuntimeError(".url_for() is not supported "
+                           "by sub-application root")
 
     def url(self, **kwargs):
+        """Construct url for route with additional params."""
+        raise RuntimeError(".url() is not supported "
+                           "by sub-application root")
+
+    def get_info(self):
+        return {'app': self._app,
+                'prefix': self._prefix}
+
+    @asyncio.coroutine
+    def resolve(self, request):
+        if not request.url.raw_path.startswith(self._prefix):
+            return None, set()
+        match_info = yield from self._app.router.resolve(request)
+        match_info.add_app(self._app)
+        if isinstance(match_info.http_exception, HTTPMethodNotAllowed):
+            methods = match_info.http_exception.allowed_methods
+        else:
+            methods = set()
+        return (match_info, methods)
+
+    def __len__(self):
+        return len(self._app.router.routes())
+
+    def __iter__(self):
+        return iter(self._app.router.routes())
+
+    def __repr__(self):
+        return "<PrefixedSubAppResource {prefix} -> {app!r}>".format(
+            prefix=self._prefix, app=self._app)
+
+
+class ResourceRoute(AbstractRoute):
+    """A route with resource"""
+
+    def __init__(self, method, handler, resource, *,
+                 expect_handler=None):
+        super().__init__(method, handler, expect_handler=expect_handler,
+                         resource=resource)
+
+    def __repr__(self):
+        return "<ResourceRoute [{method}] {resource} -> {handler!r}".format(
+            method=self.method, resource=self._resource,
+            handler=self.handler)
+
+    @property
+    def name(self):
+        return self._resource.name
+
+    def url_for(self, *args, **kwargs):
+        """Construct url for route with additional params."""
+        return self._resource.url_for(*args, **kwargs)
+
+    def url(self, **kwargs):
+        """Construct url for route with additional params."""
+        super().url(**kwargs)
+        return self._resource.url(**kwargs)
+
+    def get_info(self):
+        return self._resource.get_info()
+
+
+class SystemRoute(AbstractRoute):
+
+    def __init__(self, http_exception):
+        super().__init__(hdrs.METH_ANY, self._handler)
+        self._http_exception = http_exception
+
+    def url_for(self, *args, **kwargs):
+        raise RuntimeError(".url_for() is not allowed for SystemRoute")
+
+    def url(self, *args, **kwargs):
         raise RuntimeError(".url() is not allowed for SystemRoute")
 
-    def match(self, path):
+    @property
+    def name(self):
         return None
 
     def get_info(self):
@@ -639,26 +694,26 @@ class RoutesView(Sized, Iterable, Container):
 
 class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
 
-    DYN = re.compile(r'^\{(?P<var>[a-zA-Z][_a-zA-Z0-9]*)\}$')
+    DYN = re.compile(r'\{(?P<var>[a-zA-Z][_a-zA-Z0-9]*)\}')
     DYN_WITH_RE = re.compile(
-        r'^\{(?P<var>[a-zA-Z][_a-zA-Z0-9]*):(?P<re>.+)\}$')
+        r'\{(?P<var>[a-zA-Z][_a-zA-Z0-9]*):(?P<re>.+)\}')
     GOOD = r'[^{}/]+'
     ROUTE_RE = re.compile(r'(\{[_a-zA-Z][^{}]*(?:\{[^{}]*\}[^{}]*)*\})')
     NAME_SPLIT_RE = re.compile('[.:-]')
 
-    def __init__(self):
+    def __init__(self, app):
         super().__init__()
         self._resources = []
         self._named_resources = {}
+        self._app = app
 
     @asyncio.coroutine
     def resolve(self, request):
-        path = request.raw_path
         method = request.method
         allowed_methods = set()
 
         for resource in self._resources:
-            match_dict, allowed = yield from resource.resolve(method, path)
+            match_dict, allowed = yield from resource.resolve(request)
             if match_dict is not None:
                 return match_dict
             else:
@@ -691,20 +746,13 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
     def named_resources(self):
         return MappingProxyType(self._named_resources)
 
-    def named_routes(self):
-        # NB: it's ambiguous but it's really resources.
-        warnings.warn("Use .named_resources instead", DeprecationWarning)
-        return self.named_resources()
-
-    def register_route(self, route):
-        warnings.warn("Use resource-based interface", DeprecationWarning)
-        resource = ResourceAdapter(route)
-        self._reg_resource(resource)
-
     def _reg_resource(self, resource):
         assert isinstance(resource, AbstractResource), \
             'Instance of AbstractResource class is required, got {!r}'.format(
                 resource)
+        if self.frozen:
+            raise RuntimeError("Cannot register a resource into "
+                               "frozen router.")
 
         name = resource.name
 
@@ -727,20 +775,20 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
         if not path.startswith('/'):
             raise ValueError("path should be started with /")
         if not ('{' in path or '}' in path or self.ROUTE_RE.search(path)):
-            resource = PlainResource(path, name=name)
+            resource = PlainResource(quote(path, safe='/'), name=name)
             self._reg_resource(resource)
             return resource
 
         pattern = ''
         formatter = ''
         for part in self.ROUTE_RE.split(path):
-            match = self.DYN.match(part)
+            match = self.DYN.fullmatch(part)
             if match:
                 pattern += '(?P<{}>{})'.format(match.group('var'), self.GOOD)
                 formatter += '{' + match.group('var') + '}'
                 continue
 
-            match = self.DYN_WITH_RE.match(part)
+            match = self.DYN_WITH_RE.fullmatch(part)
             if match:
                 pattern += '(?P<{var}>{re})'.format(**match.groupdict())
                 formatter += '{' + match.group('var') + '}'
@@ -749,11 +797,12 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
             if '{' in part or '}' in part:
                 raise ValueError("Invalid path '{}'['{}']".format(path, part))
 
+            part = quote(part, safe='/')
             formatter += part
             pattern += re.escape(part)
 
         try:
-            compiled = re.compile('^' + pattern + '$')
+            compiled = re.compile(pattern)
         except re.error as exc:
             raise ValueError(
                 "Bad pattern '{}': {}".format(pattern, exc)) from None
@@ -769,23 +818,26 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
 
     def add_static(self, prefix, path, *, name=None, expect_handler=None,
                    chunk_size=256*1024, response_factory=StreamResponse,
-                   show_index=False):
+                   show_index=False, follow_symlinks=False):
         """Add static files view.
 
         prefix - url prefix
         path - folder with files
 
         """
+        # TODO: implement via PrefixedResource, not ResourceAdapter
         assert prefix.startswith('/')
         if not prefix.endswith('/'):
             prefix += '/'
-        route = StaticRoute(name, prefix, path,
-                            expect_handler=expect_handler,
-                            chunk_size=chunk_size,
-                            response_factory=response_factory,
-                            show_index=show_index)
-        self.register_route(route)
-        return route
+        resource = StaticResource(prefix, path,
+                                  name=name,
+                                  expect_handler=expect_handler,
+                                  chunk_size=chunk_size,
+                                  response_factory=response_factory,
+                                  show_index=show_index,
+                                  follow_symlinks=follow_symlinks)
+        self._reg_resource(resource)
+        return resource
 
     def add_head(self, *args, **kwargs):
         """
@@ -822,3 +874,15 @@ class UrlDispatcher(AbstractRouter, collections.abc.Mapping):
         Shortcut for add_route with method DELETE
         """
         return self.add_route(hdrs.METH_DELETE, *args, **kwargs)
+
+    def add_subapp(self, prefix, subapp):
+        assert prefix.startswith('/')
+        if not prefix.endswith('/'):
+            prefix += '/'
+        if subapp.frozen:
+            raise RuntimeError("Cannod add frozen application")
+        resource = PrefixedSubAppResource(prefix, subapp)
+        self._reg_resource(resource)
+        self._app._reg_subapp_signals(subapp)
+        subapp.freeze()
+        return resource
