@@ -1,7 +1,6 @@
 """Async gunicorn worker for aiohttp.web"""
 
 import asyncio
-import logging
 import os
 import re
 import signal
@@ -48,14 +47,8 @@ class GunicornWebWorker(base.Worker):
         sys.exit(self.exit_code)
 
     def make_handler(self, app):
-        if hasattr(self.cfg, 'debug'):
-            is_debug = self.cfg.debug
-        else:
-            is_debug = self.log.loglevel == logging.DEBUG
-
         return app.make_handler(
             logger=self.log,
-            debug=is_debug,
             slow_request_timeout=self.cfg.timeout,
             keepalive_timeout=self.cfg.keepalive,
             access_log=self.log.access_log,
