@@ -7,7 +7,7 @@ async def fetch(session):
     async with session.get(
             'http://httpbin.org/basic-auth/andrew/password') as resp:
         print(resp.status)
-        body = yield from resp.text()
+        body = await resp.text()
         print(body)
 
 
@@ -15,8 +15,8 @@ async def go(loop):
     async with aiohttp.ClientSession(
             auth=aiohttp.BasicAuth('andrew', 'password'),
             loop=loop) as session:
-        await go(session)
+        await fetch(session)
 
 
 loop = asyncio.get_event_loop()
-loop.run_until_complete(go())
+loop.run_until_complete(go(loop))
